/************************************************************************/
/* backend.c - DNS backend connection management			*/
/************************************************************************/
/*
 *  Copyright 2021-2023 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cutil/cutil.h>

#include <logging.h>
#include <udp.h>
#include <tcp.h>
#include <pfm.h>
#include <metric.h>
#include <helpers.h>
#include <pfm_dns.h>
#include <backend.h>
#include <fdfuncs.h>




extern pfm_dns_t priv_data;

/*
struct _backend_t {

	backend_protocol	protocol;

	char			*dest_host;
	char			*dest_service;
	struct sockaddr		dest_sa;

	int			conn_max;
	int			conn_count;
	int			*conn_fd_table;
};
*/


char *backend_protocol_name[] = {
	"udp",
	"tcp",
};


// forward function declarations
int backend_timer_reconnect_callback(pfm_timer_t *timer, void *arg);



int tcp_connect_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int			rc;
	socklen_t		optlen;
	int			opt;
	backend_t		*be;
	ioctl_opt_register_fd_t	o_rfd;
	pfm_timer_t		*timer;
	fd_t			*fdte;

	be = arg;

	// No matter what, unregister the FD
	o_rfd.port = 0;
	o_rfd.callback_arg = NULL;
	o_rfd.callback = tcp_connect_io_callback;
	o_rfd.fd = fd;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_UNREGISTER_BACKEND_FD,
			&o_rfd);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() pfm_vector_ioctl(fd=%d, PFM_IOCTL_UNREGISTER_BACKEND_FD)"
		" rc=%d",
		__func__,
		fd,
		rc);

	// Check the connect() result
	optlen = sizeof(opt);
	rc = getsockopt(fd,
			SOL_SOCKET,
			SO_ERROR,
			&opt,
			&optlen);
	if (rc) {
		rc = errno;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() getsockopt(fd=%d) failed rc=%d",
			__func__,
			fd,
			rc);
		fdte = &priv_data.fd_table[fd];
		if (fdte->ptr.request_table)
			rc = request_table_free(fdte->ptr.request_table);
		rc = tcp_shutdown(fd);
		// mark the backend as failed and schedule retry
		memset(&priv_data.fd_table[fd], 0, sizeof(fd_t));
		be->flags &= ~(BACKEND_F_RECONNECT_PENDING);
		if (!(be->flags & BACKEND_F_TIMER_PENDING)) {
			timer = pfm_vector_schedule_timer(priv_data.module,
				be->timer_backoff_ms,
				backend_timer_reconnect_callback,
				be);
			if (!timer) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"%s() failed to schedule reconnect"
					" timer",
					__func__);
			} else {
				be->flags |= BACKEND_F_TIMER_PENDING;
			}
		}
		return 0;
	}

	if (opt == 0) {
		// reregister FD with ioloop (w/ new callback)
		o_rfd.port = 0;
		o_rfd.callback_arg = NULL;
		o_rfd.callback = backend_io_callback;
		o_rfd.fd = fd;
		rc = pfm_vector_ioctl(priv_data.module,
				PFM_IOCTL_REGISTER_BACKEND_FD,
				&o_rfd);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() pfm_vector_ioctl(fd=%d, PFM_IOCTL_REGISTER_BACKEND_FD)"
			" rc=%d",
			__func__,
			fd,
			rc);
		// TODO check rc

		pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"%s() connect() successful fd=%d",
			__func__,
			fd);
		// reset reconnect backoff
		be->timer_backoff_ms = BACKEND_RECONNECT_BACKOFF_BASE;
		be->flags &= ~(BACKEND_F_RECONNECT_PENDING);
		// TODO check for queued backend requests and handle them
	} else {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"%s() connect() failed fd=%d rc=%d(%s), closing",
			__func__,
			fd,
			opt,
			strerror(opt));
		fdte = &priv_data.fd_table[fd];
		if (fdte->ptr.request_table)
			rc = request_table_free(fdte->ptr.request_table);
		rc = tcp_shutdown(fd);
		// mark the backend as failed and scheduled retry
		memset(&priv_data.fd_table[fd], 0, sizeof(fd_t));
		be->flags &= ~(BACKEND_F_RECONNECT_PENDING);
		if (!(be->flags & BACKEND_F_TIMER_PENDING)) {
			timer = pfm_vector_schedule_timer(priv_data.module,
				be->timer_backoff_ms,
				backend_timer_reconnect_callback,
				be);
			if (!timer) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"%s() failed to schedule reconnect"
					" timer",
					__func__);
			} else {
				be->flags |= BACKEND_F_TIMER_PENDING;
			}
		}
	}

	return 0;
}


backend_t *backend_new(backend_protocol protocol,
			char *hostname,
			char *servicename,
			int min_connections,
			int max_connections)
{
	int			rc;
	backend_t		*be;
	struct addrinfo		hint;
	struct sockaddr_in	*sa_in;
	struct addrinfo		*res;


	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() memory allocation line=%d\n",
		__func__,
		__LINE__);
	be = calloc(1, sizeof(backend_t));
	if (!be) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() memory allocation failed\n",
			__func__);
		return NULL;
	}

	be->timer_backoff_ms = BACKEND_RECONNECT_BACKOFF_BASE;
	be->conn_max = max_connections;
	be->conn_count = 0;
	be->conn_min = min_connections;
	metric_init(&be->metric[METRIC_BE_REQUESTS],
		"requests",
		METRIC_TYPE_UINT32);
	be->rate_last_time = time(NULL);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() memory allocation line=%d\n",
		__func__,
		__LINE__);
	be->conn_fd_table = calloc(be->conn_max, sizeof(int));
	if (!be->conn_fd_table) {
		pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() memory free line=%d\n",
			__func__,
			__LINE__);
		free(be);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() memory allocation failed\n",
			__func__);
		return NULL;
	}

	be->protocol = protocol;
	be->dest_host = strdup(hostname);
	be->dest_service = strdup(servicename);

	memset(&hint, 0, sizeof(struct addrinfo));
	hint.ai_family = AF_INET;
	if (be->protocol == BACKEND_PROTO_UDP) {
		hint.ai_socktype = SOCK_DGRAM;
		hint.ai_protocol = IPPROTO_UDP;
	} else if (be->protocol == BACKEND_PROTO_TCP) {
		hint.ai_socktype = SOCK_STREAM;
		hint.ai_protocol = IPPROTO_TCP;
	}
	rc = getaddrinfo(be->dest_host,
			be->dest_service,
			&hint,
			&res);
	if (rc != 0) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s()  getaddrinfo() failed rc=%d",
			__func__,
			rc);
		pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() memory free line=%d\n",
			__func__,
			__LINE__);
		free(be->dest_host);
		free(be->dest_service);
		free(be->conn_fd_table);
		free(be);
		return  NULL;
	}
	sa_in = (struct sockaddr_in *)res->ai_addr;
	memcpy(&be->dest_sa, sa_in, sizeof(struct sockaddr_in));
	freeaddrinfo(res);

	return be;
}


int backend_free(backend_t *be)
{
	(void)backend_decrease_connections(be, be->conn_count);

	pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() memory free line=%d\n",
		__func__,
		__LINE__);
	free(be->dest_host);
	free(be->dest_service);
	free(be->conn_fd_table);
	free(be);

	return 0;
}


int backend_add_connection(backend_t *be)
{
	int			rc;
	int			fd;
	ioctl_opt_register_fd_t	opt;
	ioctl_opt_modify_fd_t	o_mfd;
	fd_t			*fdte;
	request_table_t		*rt;

	if (be->conn_count >= be->conn_max)
		return E2BIG;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Adding connection to backend %s:%s:%s",
		backend_protocol_to_name(be->protocol),
		be->dest_host,
		be->dest_service);

	fd = -1;
	opt.port = 0;
	opt.callback_arg = NULL;
	opt.callback = backend_io_callback;

	rt = request_table_new(65536);
	if(!rt) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() failed to allocate a request table",
			__func__);
		return ENOMEM;
	}

	if (be->protocol == BACKEND_PROTO_UDP) {
		fd = udp_connect(be->dest_host, be->dest_service);
	} else if (be->protocol == BACKEND_PROTO_TCP) {
		fd = tcp_connect(be->dest_host, be->dest_service);
	}
	if (fd == -1) {
		rc = errno;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() failed to make backend socket connection rc=%d",
			__func__,
			rc);
		(void)request_table_free(rt);
		return rc;
	}

	fdte = &priv_data.fd_table[fd];
	fdte->backend = be;

	if (be->protocol == BACKEND_PROTO_UDP) {
		fd_entry_init(fdte,
			fd,
			FD_CONN_TYPE_UDP,
			FD_USAGE_TYPE_BACKEND,
			rt);
	}
	if (be->protocol == BACKEND_PROTO_TCP) {
		fd_entry_init(fdte,
			fd,
			FD_CONN_TYPE_TCP,
			FD_USAGE_TYPE_BACKEND,
			rt);
		if (errno == EINPROGRESS) {
			// TCP: need to wait for the socket
			// to become writable and check status
			be->flags |= BACKEND_F_RECONNECT_PENDING;
			opt.fd = fd;
			opt.callback_arg = be;
			opt.callback = tcp_connect_io_callback;
			rc = pfm_vector_ioctl(priv_data.module,
					PFM_IOCTL_REGISTER_BACKEND_FD,
					&opt);
			// TODO check rc
			o_mfd.fd = fd;
			o_mfd.flags = PFM_FD_MODIFY_F_WRITABLE;
			rc = pfm_vector_ioctl(priv_data.module,
					PFM_IOCTL_MODIFY_BACKEND_FD,
					&o_mfd);
			// TODO check rc
			// delay be->conn_fd_table update until callback
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"TCP connect in progress for backend %p"
				" fd=%d",
				be,
				fd);
			// go ahead and setup the fd so requests can be
			// queued on it
			be->conn_fd_table[be->conn_count] = fd;
			be->conn_count++;
			return 0;
		}
	}
	opt.fd = fd;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_REGISTER_BACKEND_FD,
			&opt);
	if (rc != 0) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Connection creation failed to setup"
			" fd for backend %p rc=%d",
			be,
			rc);
		if (be->protocol == BACKEND_PROTO_UDP) {
			rc = udp_shutdown(fd);
		} else if (be->protocol == BACKEND_PROTO_TCP) {
			rc = tcp_shutdown(fd);
		}
		(void)request_table_free(rt);
		return rc;
	}
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"Connection successfully added for backend=%p"
		" fd=%d fdte=%p",
		be,
		fd,
		fdte);
	be->conn_fd_table[be->conn_count] = fd;
	be->conn_count++;

	return 0;
}


int backend_increase_connections(backend_t *be, int count)
{
	int			rc;
	int			i;

	if ((be->conn_count+count) > be->conn_max)
		count = be->conn_max - be->conn_count;
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Connection increase requested for backend %p count=%d",
		be,
		count);
	if (be->conn_count >= be->conn_min)
		return 0;
	if (be->flags & BACKEND_F_RECONNECT_PENDING) // hold addition
		return 0;
	count = be->conn_min - be->conn_count;
	for (i=0 ; i<count ; i++) {
		rc = backend_add_connection(be);
		if(rc) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() failed to add a new connection rc=%d",
				__func__,
				rc);
			continue;
		}
	}
	return 0;
}


int backend_decrease_connections(backend_t *be, int count)
{
	int			rc;
	int			i;
	int			last;
	ioctl_opt_register_fd_t	opt;
	fd_t			*fdte;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Connection decrease requested for backend %p count=%d",
		be,
		count);

	if (count > be->conn_count)
		return E2BIG;

	for (i=0 ; i<count ; i++) {
		last = be->conn_count-1;
		fdte = &priv_data.fd_table[be->conn_fd_table[last]];
		if (fdte->in_pb || fdte->out_pb) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"FD for backend %p is in use (not closing)"
				" fd=%d",
				be,
				be->conn_fd_table[last]);
			return 0;
		}
		opt.fd = be->conn_fd_table[last];
		opt.port = 0;
		opt.callback = backend_io_callback;
		opt.callback_arg = NULL;
		rc = pfm_vector_ioctl(priv_data.module,
				PFM_IOCTL_UNREGISTER_BACKEND_FD,
				&opt);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to unregister from backend %p"
				" fd=%d rc=%d",
				be,
				be->conn_fd_table[last],
				rc);
			return rc;
		}
		if (be->protocol == BACKEND_PROTO_UDP) {
			rc = udp_shutdown(be->conn_fd_table[last]);
		} else if (be->protocol == BACKEND_PROTO_TCP) {
			rc = tcp_shutdown(be->conn_fd_table[last]);
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"Connection shutdown() for backend %p fd=%d rc=%d",
			be,
			be->conn_fd_table[last],
			rc);
		(void)request_table_free(fdte->ptr.request_table);
		fd_entry_clear(fdte);
		be->conn_fd_table[last] = -1;
		be->conn_count--;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"Connection closed for backend %p oset=%d",
			be,
			last);
	}
	return 0;
}


int backend_replace_connection(backend_t *be, int fd)
{
	int			rc;
	int			i;
	int			l;
	ioctl_opt_register_fd_t	opt;
	fd_t			*fdte;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s( Replacement connection requested for backend %s:%s"
		" old fd=%d",
		__func__,
		be->dest_host,
		be->dest_service,
		fd);

	// Find old fd in table
	for (i=0 ; i<be->conn_count ; i++) {
		if (be->conn_fd_table[i] == fd)
			break;
	}
	if (i == be->conn_count) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_EMERG,
			"%s() fd=%d not found in connection"
			" table, exiting",
			__func__,
			fd);
		exit(ENOENT);
		return ENOENT;
	}

	// Close the old connection
	opt.fd = be->conn_fd_table[i];
	opt.port = 0;
	opt.callback = backend_io_callback;
	opt.callback_arg = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_UNREGISTER_BACKEND_FD,
			&opt);
	if (rc != 0 && rc != ENOENT) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_EMERG,
			"%s() fd=%d failed to unregister from"
			" ioloop rc=%d, exiting",
			__func__,
			fd,
			rc);
		exit(rc);
		return rc;
	}
	if (be->protocol == BACKEND_PROTO_UDP) {
		(void)udp_shutdown(be->conn_fd_table[i]);
	} else if (be->protocol == BACKEND_PROTO_TCP) {
		(void)tcp_shutdown(be->conn_fd_table[i]);
	}

	fdte = &priv_data.fd_table[be->conn_fd_table[i]];
	(void)request_table_free(fdte->ptr.request_table);
	fd_entry_clear(fdte);

	// shift remaining connections down in the table
	for (l=i ; l<be->conn_count ; l++) {
		be->conn_fd_table[l] = be->conn_fd_table[l+1];
	}
	be->conn_count--;

	// Skip creating a new connection if the backend
	// already has enough existing connections
	if (be->conn_count >= be->conn_min) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() skip new connection creation (count>=min)",
			__func__);
		return 0;
	}

	// Create the replacement connection
	rc = backend_add_connection(be);
	if(rc) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() failed to add a new connection rc=%d",
			__func__,
			rc);
		return rc;
	}

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Replacement connection complete for backend %s:%s",
		be->dest_host,
		be->dest_service);

	return 0;
}



backend_protocol backend_name_to_protocol(char *name)
{
	int	i;
	for (i=0 ; i<_BACKEND_PROTO_MAX; i++) {
		if (!strcmp(backend_protocol_name[i], name))
			return i;
	}
	return _BACKEND_PROTO_MAX;
}


char *backend_protocol_to_name(backend_protocol protocol)
{
	if (protocol < 0 || protocol >= _BACKEND_PROTO_MAX)
		return NULL;
	return backend_protocol_name[protocol];
}



int backend_timer_reconnect_callback(pfm_timer_t *timer, void *arg)
{
	int		count;
	backend_t	*be;

	be = arg;
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() entry backend=%p min=%d count=%d max=%d flags=%d",
		__func__,
		be,
		be->conn_min,
		be->conn_count,
		be->conn_max,
		be->flags);
	be->flags &= ~(BACKEND_F_TIMER_PENDING);
	if (be->flags & BACKEND_F_RECONNECT_PENDING)
		return 0;
	if (be->conn_count >= be->conn_min)
		return 0;
	count = be->conn_min - be->conn_count;
	if (be->timer_backoff_ms < BACKEND_RECONNECT_BACKOFF_MAX)
		be->timer_backoff_ms *= 2;
	(void)backend_increase_connections(be, count);
	return 0;
}


void backend_count_usage(backend_t *be)
{
	uint32_t	c;
	uint32_t	r;
	time_t		t;
	metric_incr_u32(&be->metric[METRIC_BE_REQUESTS]);
	c = metric_get_u32(&be->metric[METRIC_BE_REQUESTS])
			 - be->rate_last_count;
	t = time(NULL) - be->rate_last_time;

	// check every 10 requests or after 5 seconds
	if (c < 10 && t < 5)
		return;
	if (be->conn_count == 0)
		return;
	be->rate_last_count += c;
	be->rate_last_time += t;
	if (t == 0)
		t = 1; //prevent div by zero

	// rate calculations for backend scaling
	// requests per backend per second (integer should be fine)
	r = c/be->conn_count/t;
//	pfm_vector_log(priv_data.module, LOG_LEVEL_NOTICE,
//		"%s() r=%d", __func__, r);
	if (r >= 10) {
		backend_add_connection(be);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"Scaled backend=%p up one connection",
			be);
	} else if (r < 1) {
		if (be->conn_count > 1) {
			backend_decrease_connections(be, 1);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Requested to scale backend=%p down one connection",
				be);
		}
	}
	return;
}


