/************************************************************************/
/* fdfuncs.c - fd_t manipulation functions				*/
/************************************************************************/
/*
 *  Copyright 2021-2023 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <assert.h>
#include <dlist.h>
#include <pfm_dns.h>
#include <fdfuncs.h>


extern pfm_dns_t priv_data;


void fd_entry_init(fd_t *fde,
		int fd,
		fd_conn_type conntype,
		fd_usage_type usagetype,
		void *dataptr)
{
	assert(fde);
	fde->fd = fd;
	fde->conntype = conntype;
	fde->usagetype = usagetype;
	fde->ptr.ptr = dataptr;
	fde->in_pb = NULL;
	fde->in_rem = 0;
	fde->out_pb = NULL;
	dlist_init(&fde->in_qhead);
	dlist_init(&fde->out_qhead);
	return;
}


void fd_entry_clear(fd_t *fde)
{
	assert(fde);
	fde->fd = -1;
	fde->conntype = 0;
	fde->usagetype = 0;
	fde->ptr.ptr = NULL;
	fde->backend = NULL;
	fde->in_pb = NULL;
	fde->in_rem = 0;
	fde->out_pb = NULL;
	return;
}


int fd_try_write(fd_t *fde)
{
	int			rc;
	uint16_t		u16;
	int			len;
	ioctl_opt_modify_fd_t	opt;
	char			*p;

	assert(fde);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fde=%p fde->fd=%d",
		__func__,
		fde,
		fde->fd);
	while (1) {
		if (!fde->out_pb) {
			// setup the next outbound packet
			if (dlist_is_empty(&fde->out_qhead)) {
				// nothing else to send
				// disable write notification
				opt.fd = fde->fd;
				opt.flags = 0;
				rc = pfm_vector_ioctl(priv_data.module,
						PFM_IOCTL_MODIFY_BACKEND_FD,
						&opt);
				return 0;
			}
			fde->out_pb = (packetbuf_t*)(&fde->out_qhead)->next;
			dlist_remove((dlist_t*)fde->out_pb);
		}

		pnetbuffer_get_readable(fde->out_pb->nb, &p, &u16);
		if (fde->conntype == FD_CONN_TYPE_TCP) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() calling send() fde=%p fde->fd=%d p=%p len=%d",
				__func__,
				fde,
				fde->fd,
				p,
				u16);
			rc = send(fde->fd,
				p,
				u16,
				0);
		} else {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() calling sendto() fde=%p fde->fd=%d p=%p len=%d salen=%d",
				__func__,
				fde,
				fde->fd,
				p,
				u16,
				fde->out_pb->saddr_len);
			rc = sendto(fde->fd,
				p,
				u16,
				0,
				&fde->out_pb->saddr,
				fde->out_pb->saddr_len);
		}
		if (rc == -1) {
			rc = errno;
			if (rc == EAGAIN || rc == EWOULDBLOCK) {
				// enable write notification
				opt.fd = fde->fd;
				opt.flags = PFM_FD_MODIFY_F_WRITABLE;
				rc = pfm_vector_ioctl(priv_data.module,
						PFM_IOCTL_MODIFY_BACKEND_FD,
						&opt);
				return 0;
			}
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() sendto() returned rc=%d fde=%p fde->fd=%d",
				__func__,
				rc,
				fde,
				fde->fd);
			return rc; // some other error occurred
		} else if (rc == 0) {
			// EOF??? this shouldn't happen
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() sendto() returned 0 fde=%p fde->fd=%d",
				__func__,
				fde,
				fde->fd);
			assert(0);
		}
		// some data was sent
		len = rc;
		rc = pnetbuffer_read_advance(fde->out_pb->nb, len);
		assert(rc == 0);
		if (len == u16) {
			// entire packet was written
			packetbuf_slab_return(priv_data.packetbuf_slab, fde->out_pb);
			fde->out_pb = NULL;
		}
	}

	return 0;
}



int split_packetbuf_to_tcp_dns_packets(fd_t *fde, packetbuf_t *pb)
{
	int		rc;
	uint16_t	len;
	uint16_t	u16;
	char		*p;
	packetbuf_t	*npb;

	assert(fde);
	assert(pb);
	while (1) {
		// attempt to read TCP length header
		rc = pnetbuffer_peek_uint16(pb->nb, &len);
		if (rc)
			return EAGAIN;
		// see if the whole packet is here
		pnetbuffer_get_readable(pb->nb, &p, &u16);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() pb readable bytes=%d",
			__func__,
			u16);
		if (u16 < (len+2))
			break;
		// move whole packet to new packetbuf and queue it
		npb = packetbuf_slab_get(priv_data.packetbuf_slab);
		assert(npb);
		pnetbuffer_add_data(npb->nb, p, len+2);
		dlist_add_tail(&fde->in_qhead, (dlist_t*)npb);
		pnetbuffer_read_advance(pb->nb, len+2);
		if (u16 == (len+2)) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() returning with empty pnetbuffer",
				__func__);
			return 0; // no more data in packetbuf (caller should free it)
		}
	}

	// Still some data, slide packet data down
	// NOTE treating pnetbuffer this way is a bit of a hack
	pnetbuffer_get_readable(pb->nb, &p, &u16);
	pnetbuffer_reset(pb->nb);
	pnetbuffer_add_data(pb->nb, p, u16);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() slid bytes to bottom of pnetbuffer bytes=%d",
		__func__,
		u16);
	return EAGAIN;
}


int fd_try_read(fd_t *fde)
{
	int			rc;
	uint16_t		u16;
	int			len;
	char			*p;
//	packetbuf_t		*pb;

	assert(fde);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fde=%p fde->fd=%d",
		__func__,
		fde,
		fde->fd);
	while (1) {
		if (!fde->in_pb) {
			fde->in_pb = packetbuf_slab_get(priv_data.packetbuf_slab);
			assert(fde->in_pb);
//			if (fde->conntype == FD_CONN_TYPE_TCP)
//				fde->in_rem = 0; // reset remaining bytes
		}
		fde->in_pb->saddr_len = sizeof(struct sockaddr);
		pnetbuffer_get_writable(fde->in_pb->nb, &p, &u16);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() calling recvfrom() fde=%p fde->fd=%d p=%p len=%d salen=%d",
			__func__,
			fde,
			fde->fd,
			p,
			u16,
			fde->in_pb->saddr_len);
		rc = recvfrom(fde->fd,
			p,
			(int)u16,
			0,
			&fde->in_pb->saddr,
			&fde->in_pb->saddr_len);
		if (rc == -1) {
			rc = errno;
			if (rc == EAGAIN || rc == EWOULDBLOCK)
				return 0;
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() recvfrom() returned rc=%d fde=%p fde->fd=%d",
				__func__,
				rc,
				fde,
				fde->fd);
			return rc;
		} else if (rc == 0) {
			// EOF on socket
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() recvfrom() returned 0 fde=%p fde->fd=%d",
				__func__,
				fde,
				fde->fd);
			return EBADFD; // TCP socket closed
		}
		// some data was read
		len = rc;
		rc = pnetbuffer_set_written(fde->in_pb->nb, len);
		assert(rc == 0);
		if (fde->conntype == FD_CONN_TYPE_TCP) {
			// TCP packets have a length header and may be
			// split aross reads
			rc = split_packetbuf_to_tcp_dns_packets(fde, fde->in_pb);
			if (rc == 0) {
				// no data left after processing, free it
				// (or maybe just reuse it?)
				packetbuf_slab_return(priv_data.packetbuf_slab, fde->in_pb);
				fde->in_pb = NULL;
//				fde->in_rem = 0;
			}

//			// determine if any data still needs to be read
//			// for this DNS packet
//			if (fde->in_rem == 0) {
//				// beginning of a TCP DNS packet
//				// get packet length
//				rc = pnetbuffer_read_uint16(fde->in_pb->nb, &u16);
//				assert(rc == 0);
//				fde->in_rem = u16;
//				len -= 2;
//			}
//			pnetbuffer_get_readable(fde->in_pb->nb, &p, &u16);
//			if (u16 >= fde->in_rem) {
//				// finished reading a packet
//				pnetbuffer_reset_read(fde->in_pb->nb);
//				if (u16 == fde->in_rem) {
//					// exact length
//					dlist_add_tail(&fde->in_qhead, (dlist_t*)fde->in_pb);
//					fde->in_pb = NULL;
//					fde->in_rem = 0;
//				} else {
//					// more than one packet's data
//					// copy the complete packet to new packetbuf
//					// TODO this isn't too efficient, rework it
//					pb = packetbuf_slab_get(priv_data.packetbuf_slab);
//					assert(pb);
//					pnetbuffer_get_readable(fde->out_pb->nb, &p, &u16);
//					pnetbuffer_add_data(pb->nb, p, fde->in_rem+2);
//					pnetbuffer_read_advance(fde->out_pb->nb, fde->in_rem+2);
//					dlist_add_tail(&fde->in_qhead, (dlist_t*)pb);
//					// copy remaining to new packetbuf
//					pb = packetbuf_slab_get(priv_data.packetbuf_slab);
//					assert(pb);
//					pnetbuffer_get_readable(fde->out_pb->nb, &p, &u16);
//					pnetbuffer_add_data(pb->nb, p, u16);
//					// free original packetbuf and setup for next TCP DNS header
//					packetbuf_slab_return(priv_data.packetbuf_slab, fde->in_pb);
//					fde->in_pb = pb;
//					fde->in_rem = 0;
//				}
//			}
		} else {
			// UDP packets are all-or-nothing
			dlist_add_tail(&fde->in_qhead, (dlist_t*)fde->in_pb);
			fde->in_pb = NULL;
		}
	}

	return 0;
}


// submit a packetbuf_t for writing to the socket
int fd_submit_for_write(fd_t *fde, packetbuf_t *pb)
{
	// this assumes TCP packets already have the
	// length header in the packetbuf
	assert(fde);
	assert(pb);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fde=%p fde->fd=%d pb=%p",
		__func__,
		fde,
		fde->fd,
		pb);
	if (fde->out_pb) {
		// already sending a packet, so queue this one
		dlist_add_tail(&fde->out_qhead, (dlist_t*)pb);
	} else {
		// not currently sending, so prepare to send this one
		fde->out_pb = pb;
	}
	return fd_try_write(fde);
}


// return a complete packet read from the socket
// (returns NULL if no complete packet is waiting
packetbuf_t *fd_get_ready_packet(fd_t *fde)
{
	int		rc;
	dlist_t		*item;
	assert(fde);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fde=%p fde->fd=%d",
		__func__,
		fde,
		fde->fd);
	// first try to read from the socket
	if (dlist_is_empty(&fde->in_qhead)) {
		rc = fd_try_read(fde);
		(void)rc; // prevent compiler warning
//		if (rc != EAGAIN && rc != EWOULDBLOCK)
//			return NULL;
	}
	// return a complete packet, if available
	if (!dlist_is_empty(&fde->in_qhead)) {
		item = (&fde->in_qhead)->next;
		dlist_remove(item);
		return (packetbuf_t*)item;
	}
	return NULL;
}



