/************************************************************************/
/* packetbuf.c - Packetbuf module					*/
/************************************************************************/
/*
 *  (c)2021,2022,2024 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <dlist.h>
#include <pnetbuffer.h>
#include <packetbuf.h>


packetbuf_t *packetbuf_alloc_one(void)
{
	packetbuf_t		*pb;

	pb = calloc(1, sizeof(packetbuf_t));
	if (!pb) return NULL;
	pb->nb = pnetbuffer_init(pb->nb_buffer, PACKET_BUFFER_SIZE);
	if (!pb->nb) {
		free(pb);
		return NULL;
	}
	return pb;
}


packetbuf_slab_t *packetbuf_slab_alloc(uint32_t initsize, uint32_t maxsize)
{
	uint32_t		i;
	packetbuf_t		*pb;
	packetbuf_slab_t	*slab;

	slab = calloc(1, sizeof(packetbuf_slab_t));
	if (!slab) return NULL;
	slab->size_max = maxsize;
	slab->size_curr = initsize;
	dlist_init(&slab->head);

	for (i=0 ; i< initsize ; i++) {
		pb = packetbuf_alloc_one();
		if (!pb) goto packetbuf_slab_alloc_err;
		dlist_add(&slab->head, (dlist_t*)pb);
	}
	return slab;

packetbuf_slab_alloc_err:
	packetbuf_slab_free(slab); // free what was alloc'ed
	return NULL;
}


int packetbuf_slab_free(packetbuf_slab_t *slab)
{
	dlist_t			*li;
	
	assert(slab);
	if (!slab) return EINVAL;
	// We can only free the items that are still in the list.
	// Don't free the slab until all items have been returned.
	while (!dlist_is_empty(&slab->head)) {
		li = (&slab->head)->next;
		dlist_remove(li);
		free((packetbuf_t*)li);
	}
	free(slab);
	return 0;
}


packetbuf_t *packetbuf_slab_get(packetbuf_slab_t *slab)
{
	packetbuf_t	*pb;
	dlist_t		*li;

	assert(slab);
	if (!dlist_is_empty(&slab->head)) {
		// use an item from the slab
		li = (&slab->head)->next;
		dlist_remove(li);
		pnetbuffer_reset(((packetbuf_t*)li)->nb);
		return (packetbuf_t*)li;
	}
	// allocate another item (up to size_max)
	if (slab->size_curr == slab->size_max)
		return NULL;
	pb = packetbuf_alloc_one();
	if (!pb)
		return NULL;
	slab->size_curr++;
	return pb;
}


void packetbuf_slab_return(packetbuf_slab_t *slab, packetbuf_t *pb)
{
	assert(slab);
	assert(pb);
	dlist_add(&slab->head, (dlist_t*)pb);
	return;
}


void packetbuf_set_address(packetbuf_t *pb,
			struct sockaddr *sa,
			socklen_t sa_len)
{
	assert(pb);
	assert(sa);
	assert(sa_len);
	if (sa && sa_len) {
		memcpy((char*)&pb->saddr, sa, sa_len);
		pb->saddr_len = sa_len;
	}
	return;
}


void packetbuf_get_address(packetbuf_t *pb,
			struct sockaddr *sa,
			socklen_t *sa_len)
{
	assert(pb);
	assert(sa);
	assert(sa_len);
	assert(pb->saddr_len);
	assert(*sa_len >= pb->saddr_len);
	memcpy(sa, (char*)&pb->saddr, pb->saddr_len);
	*sa_len = pb->saddr_len;
	return;
}


void packetbuf_set_data(packetbuf_t *pb, void *data)
{
	assert(pb);
	pb->data = data;
	return;
}

void *packetbuf_get_data(packetbuf_t *pb)
{
	assert(pb);
	return pb->data;
}


